<?php

declare(strict_types=1);

namespace app\controllers;

use yii;
use yii\filters\AccessControl;
use yii\web\Controller;
use yii\web\Response;
use yii\web\Cookie;

use app\models\Order;
use app\models\Account;
use app\models\Product;
use app\models\OrderEdgeSupply;
use app\models\AccountEdgeOrder;
use app\models\connection\Dellin;
use app\models\Notification;
use app\models\SupplyEdgeProduct;

use Codeception\PHPUnit\ResultPrinter\HTML;

use Exception;

class OrderController extends Controller
{
    public function behaviors()
    {
        return [
            'access' => [
                'class' => AccessControl::class,
                'rules' => [
                    [
                        'allow' => true,
                        'roles' => ['@'],
                        'actions' => ['index', 'accept', 'read', 'write', 'delete', 'amount-update', 'pay', 'supply-read', 'supply-write-stts']
                    ],
                    [
                        'allow' => false,
                        'roles' => ['?'],
                        'denyCallback' => [$this, 'accessDenied']
                    ]
                ]
            ]
        ];
    }

    public function accessDenied()
    {
        // Инициализация
        $cookies = yii::$app->response->cookies;

        // Запись cookie с редиректом, который выполнится после авторизации
        $cookies->add(new Cookie([
            'name' => 'redirect',
            'value' => yii::$app->request->pathInfo
        ]));

        if (Yii::$app->request->isPost) {
            // POST-запрос

            // Настройка
            Yii::$app->response->format = Response::FORMAT_JSON;

            // Генерация ответа
            Yii::$app->response->content = json_encode([
                'main' => $this->renderPartial('/orders/index'),
                'redirect' => '/authentication',
                '_csrf' => Yii::$app->request->getCsrfToken()
            ]);
        } else if (Yii::$app->request->isGet) {
            // GET-запрос

            $this->redirect('/authentication');
        }
    }

    public function actionIndex()
    {
        // Инициализация
        $orders = Order::search(type: 'all', limit: 10, page: 1, select: '{account_edge_order, order}');
        $moderator_orders = self::genOrdersForModeration();

        if (yii::$app->request->isPost) {
            // POST-запрос

            // Настройка
            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/orders/index', compact('orders', 'moderator_orders')),
                'title' => 'Заказы',
                'redirect' => '/orders',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/orders/index', compact('orders', 'moderator_orders'));
    }

    /**
     * Подтверждение заказа
     *
     * @param int $_key Ключ записи в коллекции order
     *
     * @return string|array|null
     */
    public function actionAccept(int $_key): string|array|null
    {
        // Инициализация
        $buffer = [];
        $account_edge_order = AccountEdgeOrder::searchByOrder(Order::collectionName() . '/' . $_key, limit: 1)[0];

        // Запись в буфер записи в базу данных
        $account_edge_order->type = 'accepted';

        // Отправка изменений в базу данных
        if ($account_edge_order->save()) {
            // Удалось сохранить изменения

            // Инициализация
            $orders = Order::search(type: 'all', limit: 10, page: 1, select: '{account_edge_order, order}');
            $moderator_orders = self::genOrdersForModeration();

            // Запись в буфер вывода
            $buffer['main'] = $this->renderPartial('/orders/index', compact('orders', 'moderator_orders'));
        } else {
            // Не удалось сохранить изменения

            yii::$app->response->statusCode = 500;
        }

        // Запись обязательных параметров
        $buffer['_csrf'] = yii::$app->request->getCsrfToken();

        // Настройка ответа
        yii::$app->response->format = Response::FORMAT_JSON;

        return $buffer;
    }

    /**
     * Запись
     */
    public function actionWrite(): string|array|null
    {
        if (yii::$app->request->isPost) {
            // POST-запрос

            // Инициализация входных данных
            $account = yii::$app->user->identity;
            $supply_id = yii::$app->request->post('supply_id');
            $amount = yii::$app->request->post('amount') ?? 1;
            $delivery_type = yii::$app->request->post('delivery_type');

            // Запись настроек ответа в буфер
            yii::$app->response->format = Response::FORMAT_JSON;

            if (empty($supply_id) || empty($delivery_type)) {
                // Не выполнены условия для выполнения

                // 501 Not Implemented
                yii::$app->response->statusCode = 501;
            }

            // Инициализация возврата по умолчанию
            $return = [
                '_csrf' => yii::$app->request->getCsrfToken()
            ];

            // Инициализация корзины
            if (!$model = Order::search($account)) {
                // Корзина не найдена (текущий заказ)

                // Инициализация
                $model = new Order();
                $model->save() or throw new Exception('Не удалось инициализировать заказ');

                // Запись ребра: АККАУНТ -> ЗАКАЗ
                AccountEdgeOrder::write($account->readId(), $model->readId(), 'current') or $model->addError('errors', 'Не удалось инициализировать ребро: АККАУНТ -> ЗАКАЗ');
            }

            // Если запись не удалась, то вернуть код: 500 Internal Server Error
            $model->writeSupply($supply_id, $delivery_type, (int) $amount) or yii::$app->response->statusCode = 500;

            return $return;
        }

        yii::$app->response->statusCode = 500;

        return null;
    }

    /**
     * Удаление
     *
     * @todo Разделить логику (для удобства чтения)
     */
    public function actionDelete(): string|array|null
    {
        // Инициализация
        $targets = yii::$app->request->post('targets') ?? yii::$app->request->get('targets');
        $page = yii::$app->request->get('page') ?? yii::$app->request->post('page') ?? 1;
        $account = yii::$app->user->identity;
        $order = Order::search();

        if ($targets) {
            // Удаление выбранных целей (удаление из заказа)

            foreach (isset($targets[0]) && is_array($targets[0]) ? $targets : [$targets] as $target) {
                // Унификация входных параметров

                foreach ($target as $catn => $amount) {
                    // Перебор целей

                    // Удаление
                    $order->deleteSupply([$catn => (int) $amount]);
                }
            }
        } else {
            // Целью подразумевается сам заказ (удаление заказа)

            // Инициализация
            $order->stts = 'reserved';

            if ($order->update()) {
                // Запись в журнал
                $order->journal('reserved');

                // Поиск
                $edge = AccountEdgeOrder::searchByVertex($account->readId(), $order->readId(), 'current');

                if (count($edge) > 1) {
                    // Найден более чем 1 заказ

                    return null;
                }

                // Инициализация
                $edge = $edge[0];
                $edge->type = 'reserved';

                // Запись
                $edge->update();

                // Реинициализация
                $order = Order::search();

                if (empty($order)) {
                    // Корзина не инициализирована

                    // Инициализация
                    $order = new Order();
                    $order->save() or throw new Exception('Не удалось инициализировать заказ');

                    // Подключение
                    $order->connect($account);
                }
            }
        }

        // Инициализация содержимого корзины
        $connections = $order->content(10, $page);

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/cart/index', compact('order', 'connections')),
                'title' => 'Корзина',
                'redirect' => '/cart',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/cart/index', compact('order', 'connections'));
    }

    /**
     * Обновление количества товара
     */
    public function actionAmountUpdate(): string|array|null
    {
        // Инициализация
        $targets = yii::$app->request->post('targets') ?? yii::$app->request->get('targets');
        $page = yii::$app->request->get('page') ?? yii::$app->request->post('page') ?? 1;
        $account = yii::$app->user;
        $order = Order::search();
        $connections = $order->content(10, $page);

        foreach (isset($targets[0]) && is_array($targets[0]) ? $targets : [$targets] as $target) {
            // Унификация входных параметров

            foreach ($target as $catn => $amount) {
                // Перебор целей (переданных объектов в корзине)

                foreach ($connections as $connection) {
                    // Перебор объектов в корзине

                    if ($connection['supply']['catn'] === $catn) {
                        // Цель найдена

                        if ($connection['amount'] > $amount) {
                            // Запрошено уменьшение количества

                            // Удаление
                            $order->deleteSupply([$catn => $connection['amount'] - $amount]);
                        } else if ($connection['amount'] < $amount) {
                            // Запрошено увеличение количества

                            // Запись
                            $order->writeSupply([$connection['supply']['catn'] => $amount - $connection['amount']]);
                        }
                    }
                }
            }
        }

        // Ренициализация
        $connections = $order->content(10, $page);

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/cart/index', compact('order', 'connections')),
                'title' => 'Корзина',
                'redirect' => '/cart',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/cart/index', compact('order', 'connections'));
    }

    /**
     * Оплата
     */
    public function actionPay(): string|array|null
    {
        // Инициализация
        $model = Order::search();

        // Поиск ребра
        $edge = AccountEdgeOrder::searchByVertex(yii::$app->user->id, $model->readId(), 'current');

        if (count($edge) > 1) {
            // Найден более чем 1 заказ

            return null;
        }

        // Инициализация
        $edge = $edge[0];

        // Запись
        $edge->type = 'requested';

        if ($edge->update()) {
            // Удалось сохранить изменения

            // Запись в журнал
            $model->journal('requested');

            // Отправка уведомлений модераторам
            Notification::_write($this->renderPartial('/notification/system/newOrder', ['id' => $edge->_key]), true, '@auth', Notification::TYPE_MODERATOR_ORDER_NEW);
        }

        // Инициализация
        $orders = Order::search(type: 'all', limit: 30, page: 1, select: '{account_edge_order, order}');
        $moderator_orders = Order::search(account: '@all', type: 'requested', limit: 10, page: 1, select: '{account_edge_order, order}');

        if (yii::$app->request->isPost) {
            // POST-запрос

            yii::$app->response->format = Response::FORMAT_JSON;

            return [
                'main' => $this->renderPartial('/orders/index', compact('orders', 'moderator_orders')),
                'title' => 'Заказы',
                'redirect' => '/orders',
                '_csrf' => yii::$app->request->getCsrfToken()
            ];
        }

        return $this->render('/orders/index', compact('orders', 'moderator_orders'));
    }

    /**
     * Генерация данных заказов для модераторов
     *
     * Включает поиск запрошенных заказов и связанных с ними поставках
     *
     * @return array ['order' => array, 'order_edge_account' => array, 'supplies' => array]
     */
    protected static function genOrdersForModeration(int $page = 1): array
    {
        $orders = Order::search(account: '@all', type: 'requested', limit: 10, page: 1, select: '{account_edge_order, order}');

        foreach ($orders as &$order) {
            // Перебор заказов

            $order['supplies'] = Order::searchById($order['order']['_id'])->content(10, $page);
        }

        return $orders;
    }

    /**
     * Чтение инстанции поставки в заказе (order_edge_supply)
     *
     * @param int $_key Ключ записи в коллекции order_edge_supply
     *
     * @return string|array|null
     */
    public function actionSupplyRead(int $_key): string|array|null
    {
        // Инициализация
        $buffer = [];

        if ($order_edge_supply = OrderEdgeSupply::searchById($_id = OrderEdgeSupply::collectionName() . '/' . $_key)) {
            // Удалось найти инстанцию поставки

            // Инициализация ребра: ПОСТАВКА -> ТОВАР
            $supply_edge_product = SupplyEdgeProduct::searchBySupplyId($order_edge_supply->_to);

            // // Поиск ребра до аккаунта
            // $account = Account::searchBySupplyId($order_edge_supply->_to);

            // // Поиск привязанного товара
            // $product = Product::searchBySupplyId($order_edge_supply->_to);

            // try {
            //     // Инициализация доставки
            //     $delivery = Dellin::calcDeliveryAdvanced(
            //         explode('_', $account['opts']['delivery_from_terminal'])[1],
            //         explode('_', yii::$app->user->identity->opts['delivery_to_terminal'])[1],
            //         (int) ($product['wght'] ?? 0),
            //         (int) ($product['dmns']['x'] ?? 0),
            //         (int) ($product['dmns']['y'] ?? 0),
            //         (int) ($product['dmns']['z'] ?? 0)
            //     );
            // } catch (Exception $e) {
            //     // var_dump(json_decode($e->getMessage(), true)['errors']); die;
            // }

            // Запись параметров
            $buffer['cost'] = $order_edge_supply->cost ?? $order_edge_supply->cost = $supply_edge_product[0]['onec']['Цены']['Цена']['ЦенаЗаЕдиницу'];
            $buffer['time'] = $order_edge_supply->time ?? 0;
            // $buffer['time'] = $order_edge_supply->time ?? $delivery ?? 0;
            $buffer['comm'] = $order_edge_supply->comm ?? $order_edge_supply->comm = 'Комментарий к заказу';
            $buffer['stts'] = $order_edge_supply->stts ?? $order_edge_supply->stts = 0;

            $order_edge_supply->save();
        }

        // Запись обязательных параметров
        $buffer['_csrf'] = yii::$app->request->getCsrfToken();

        // Настройка ответа
        yii::$app->response->format = Response::FORMAT_JSON;

        return $buffer;
    }

    /**
     * Чтение инстанции поставки в заказе (order_edge_supply)
     *
     * @param int $_key Ключ записи в коллекции order_edge_supply
     *
     * @return string|array|null
     */
    public function actionSupplyWriteStts(int $_key): string|array|null
    {
        // Инициализация
        $stts = yii::$app->request->post('stts') ?? yii::$app->request->get('stts');
        $buffer = [];

        if ($order_edge_supply = OrderEdgeSupply::searchById($_id = OrderEdgeSupply::collectionName() . '/' . $_key)) {
            // Удалось найти инстанцию поставки

            // Запись в буфер вывода и буфер записи в базу данных
            $buffer['stts'] = $order_edge_supply->stts = $stts;

            // Отправка в базу данных
            if ($order_edge_supply->save()) {
                // Удалось сохранить изменения


            } else {
                // Не удалось сохранить изменения

                yii::$app->response->statusCode = 500;
            }
        }

        // Запись обязательных параметров
        $buffer['_csrf'] = yii::$app->request->getCsrfToken();

        // Настройка ответа
        yii::$app->response->format = Response::FORMAT_JSON;

        return $buffer;
    }
}
