<?php

declare(strict_types=1);

namespace app\models;

use yii;

use app\models\traits\SearchByEdge;

use carono\exchange1c\controllers\ApiController;
use carono\exchange1c\interfaces\DocumentInterface;

use app\models\connection\Dellin;

use Exception;

/**
 * Заказ
 *
 * @see Account Заказчик
 * @see Supply Поставки для заказа
 */
class Order extends Document implements DocumentInterface
{
    use SearchByEdge;

    /**
     * Поставки для записи
     */
    public array $supplies;

    /**
     * Имя коллекции
     */
    public static function collectionName(): string
    {
        return 'order';
    }

    /**
     * Свойства
     */
    public function attributes(): array
    {
        return array_merge(
            parent::attributes(),
            [
                'ocid',
                'stts',
                'sync'
            ]
        );
    }

    /**
     * Метки свойств
     */
    public function attributeLabels(): array
    {
        return array_merge(
            parent::attributeLabels(),
            [
                'ocid' => 'Идентификатор 1C',
                'stts' => 'Статус',
                'sync' => 'Статус синхронизации с 1C'
            ]
        );
    }

    /**
     * Правила
     */
    public function rules(): array
    {
        return array_merge(
            parent::rules(),
            [
                [
                    'sync',
                    'boolean',
                    'message' => '{attribute} должен иметь логическое значение'
                ],
                [
                    'sync',
                    'default',
                    'value' => false
                ]
            ]
        );
    }

    /**
     * Подключение к аккаунту
     */
    public function connect(Account $account): ?AccountEdgeOrder
    {
        // Запись ребра: АККАУНТ -> ЗАКАЗ
        return AccountEdgeOrder::write($account->readId(), $this->readId(), 'current') ?? throw new Exception('Не удалось инициализировать ребро: АККАУНТ -> ЗАКАЗ');
    }

    /**
     * Запись товара к заказу
     *
     * $supply = [ Supply $supply, int $amount = 1 ]
     *
     * @param string $supply_id Идентификатор поставки
     * @param string $delivery_type Тип доставки
     * @param int $amount Количество
     * @param Account $trgt Заказчик
     *
     * @return int Количество записанных поставок
     *
     * @todo Создать параметр разделителя для администрации
     */
    public function writeSupply(string $supply_id, string $delivery_type, int $amount = 1, Account $trgt = null): int
    {
        // Инициализация
        $trgt ?? $trgt = yii::$app->user->identity ?? throw new Exception('Не удалось инициализировать заказчика');

        // Проверка корзины
        if (is_null($this->_key)) {
            // Корзина не инициализирована

            // Инициализация корзины
            if (!$this->save()) {
                // Инициализация корзины (активного заказа) не удалась

                throw new Exception('Ошибка при записи заказа в базу данных');
            }

            // Инициализация ребра: АККАУНТ -> ЗАКАЗ
            if (!AccountEdgeOrder::write($trgt->readId(), $this->readId(), 'create')) {
                // Инициализация не удалась

                throw new Exception('Ошибка при записи ребра от аккаунта до заказа в базу данных');
            }
        }

        // Инициализация
        $amount_buffer = 0;

        // Обработка поставок
        for ($i = 0; $i < $amount; $i++) {
            // Создание рёбер соразмерно запросу (добавление нескольких продуктов в корзину)

            // Запись ребра: ЗАКАЗ -> ПОСТАВКА
            if (!$supply_model = Supply::searchById($supply_id) or !$order_edge_supply = OrderEdgeSupply::write($this->readId(), $supply_model->readId(), 'write')) {
                // Поставка не найдена или запись ребра не удалась

                continue;
            } else {
                // Ребро создано (товар подключен к заказу)

                // Обновление счётчика добавленных товаров
                $amount_buffer++;

                // Запись типа доставки
                $order_edge_supply->dlvr = [
                    'type' => $delivery_type
                ];
                $order_edge_supply->update();

                // Запись в журнал
                $this->journal('write', ['target' => $supply_model->readId()]);
            }
        }

        if ($amount_buffer === 0) {
            // Отправка уведомления
            self::notification('Неудачная попытка добавить товар в корзину');
        } else if ($amount_buffer === 1) {
            // Отправка уведомления
            self::notification('Товар ' . $supply_model->catn . ' добавлен в корзину');
        } else {
            // Отправка уведомления
            self::notification('Добавлено ' . $amount_buffer . ' товаров в корзину');
        }

        return $amount_buffer;
    }

    /**
     * Удаление поставки
     *
     * @param Supply|string|array $supply Товары
     *
     * @return int Количество удалённых рёбер
     */
    public function deleteSupply(Supply|string|array $supply): int
    {
        // Инициализация
        $amount = 0;

        if ($supply instanceof Supply) {
            // Передана инстанция класса поставки или второй элемент массива не является числом

            // Унификация входных данных
            $supply = [$supply->catn => 1];
        }

        foreach (is_array($supply) ? $supply : [$supply => 1] as $catn => $amount_raw) {
            // Перебор товаров

            if ($supply = Supply::searchByCatn($catn)) {
                foreach (OrderEdgeSupply::searchByVertex($this->readId(), $supply->readId(), limit: $amount_raw) as $edge) {
                    // Перебор рёбер до продукта (если товаров в заказе несколько)

                    // Удаление
                    $edge->delete();

                    // Запись в журнал
                    $this->journal('delete', ['target' => $supply->readId()]);

                    // Постинкрементация счётчика удалённых рёбер
                    $amount++;
                }
            }
        }

        if ($amount === 0) {
            // Отправка уведомления
            self::notification('Неудачная попытка удалить товар из корзины');
        } else if ($amount === 1) {
            // Отправка уведомления
            self::notification('Товар ' . $supply->catn . ' удалён из корзины');
        } else {
            // Отправка уведомления
            self::notification('Удалено ' . $amount . ' товаров из корзины');
        }

        return $amount;
    }

    /**
     * Поиск заказа
     */
    public static function search(Account|string $account = null, string $type = 'current', int $limit = 1, int $page = 1, string $select = null): self|array|null
    {
        // Инициализация аккаунта
        if (empty($account) && isset(yii::$app->user->identity)) {
            // Данные не переданы

            $subquery_where = [
                [
                    'account._id' =>  yii::$app->user->identity->readId()
                ]
            ];
        } else if ($account instanceof Account) {
            // Передан аккаунт

            $subquery_where = [
                [
                    'account._id' =>  $account->readId()
                ]
            ];
        } else if (str_contains($account, '@all')) {
            // Запрос на обработку всех аккаунтов

            $subquery_where = [];
        } else {
            throw new Exception('Не удалось инициализировать пользователя');
        }

        // Инициализация типа заказа
        if (strcasecmp($type, 'all') !== 0) {
            // Если не указан поиск всех заказов

            $subquery_where[] = [
                'account_edge_order.type' => $type
            ];
        }


        // Инициализация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);

        // Запрос
        $return = self::searchByEdge(
            from: 'account',
            to: 'order',
            subquery_where: $subquery_where,
            foreach: ['edge' => 'account_edge_order'],
            where: 'edge._to == order._id',
            limit: $limit,
            offset: $offset,
            sort: ['DESC'],
            select: $select,
            direction: 'INBOUND'
        );

        return $limit === 1 ? $return[0] ?? null : $return;
    }

    /**
     * Поиск содержимого заказа
     *
     * @todo В будущем возможно заказ не только поставок реализовать
     */
    public function content(int $limit = 1, int $page = 1): Supply|array|null
    {
        // Генерация сдвига по запрашиваемым данным (пагинация)
        $offset = $limit * ($page - 1);

        // Поиск по рёбрам: ЗАКАЗ -> ПОСТАВКА
        $connections = Supply::searchByEdge(
            from: 'order',
            to: 'supply',
            edge: 'order_edge_supply',
            subquery_where: [
                [
                    'order._id' => $this->readId()
                ]
            ],
            foreach: ['edge' => 'order_edge_supply'],
            where: 'edge._to == supply._id',
            limit: $limit,
            offset: $offset,
            direction: 'INBOUND',
            select: '{supply, order_edge_supply}'
        );


        // Разделение поставок по типам доставки
        // foreach ($connections as $key => &$connection) {
        //     // Перебор поставок

        //     $connection
        // }

        // Инициализация реестра дубликатов
        $registry = [];

        // Подсчёт и перестройка массива для очистки от дубликатов
        foreach ($connections as $key => &$connection) {
            // Перебор поставок

            if (in_array([
                'catn' => $connection['supply']['catn'],
                'type' => $connection['order_edge_supply'][0]['dlvr']['type'] ?? 'auto'
            ], $registry)) {
                // Если данная поставка найдена в реестре

                // Удаление
                unset($connections[$key]);

                // Пропуск итерации
                continue;
            }

            // Инициализация
            $amount = 0;

            // Повторный перебор для поиска дубликатов
            foreach ($connections as &$connection_for_check) {
                if ($connection == $connection_for_check) {
                    // Найден дубликат

                    // Постинкрементация счётчика
                    $amount++;

                    // Запись в реестр
                    $registry[] = [
                        'catn' => $connection_for_check['supply']['catn'],
                        'type' => $connection['order_edge_supply'][0]['dlvr']['type'] ?? 'auto'
                    ];
                }
            }

            // Запись количества для заказа
            $connection['amount'] = $amount;
        }

        // Инициализация дополнительных данных
        foreach ($connections as $key => &$connection) {
            // Перебор поставок

            // Чтение стоимости
            $cost = Supply::readCostById($connection['supply']['_id']);

            if (empty($cost) || $cost['ЦенаЗаЕдиницу'] < 1) {
                // Если стоимость не найдена или равна нулю (явная ошибка)

                // Удаление из базы данных
                $this->deleteSupply($connection['supply']->readId());

                // Удаление из списка
                unset($connections[$key]);

                // Пропуск итерации
                continue;
            }

            // Поиск ребра до аккаунта
            $connection['account'] = Account::searchBySupplyId($connection['supply']['_id']);

            // Поиск привязанного товара
            $connection['product'] = Product::searchBySupplyId($connection['supply']['_id']);

            if (empty($connection['order_edge_supply']['dlvr']) || $connection['order_edge_supply']['dlvr']['type'] === 'auto') {
                // Доставка автоматическая

                try {
                    // Инициализация доставки Dellin (автоматическая)
                    $connection['delivery'] = Dellin::calcDeliveryAdvanced(
                        $connection['account']['opts']['delivery_from_terminal'],
                        yii::$app->user->identity->opts['delivery_to_terminal'],
                        (int) ($connection['product']['wght'] ?? 0),
                        (int) ($connection['product']['dmns']['x'] ?? 0),
                        (int) ($connection['product']['dmns']['y'] ?? 0),
                        (int) ($connection['product']['dmns']['z'] ?? 0),
                        count($connection['order_edge_supply'])
                    );
                } catch (Exception $e) {
                    $connection['delivery']['error'] = true;

                    // var_dump($e->getMessage());
                    // var_dump($e->getTrace());
                    // var_dump($e->getFile()); die;

                    // var_dump(json_decode($e->getMessage(), true)['errors']); die;
                }

                // Запись цены (цена поставки + цена доставки + наша наценка)
                $connection['cost'] = ($cost['ЦенаЗаЕдиницу'] ?? $connection['supply']->onec['Цены']['Цена']['ЦенаЗаЕдиницу']) + ($connection['delivery']['price']['all'] ?? $connection['delivery']['price']['one'] ?? 0) + ($settings['increase'] ?? 0) ?? 0;
            } else {
                // Доставка самолётом

                try {
                    // Инициализация доставки Dellin (самолётом)
                    $connection['delivery']['avia'] = Dellin::calcDeliveryAdvanced(
                        $connection['account']['opts']['delivery_from_terminal'],
                        yii::$app->user->identity->opts['delivery_to_terminal'],
                        (int) ($connection['product']['wght'] ?? 0),
                        (int) ($connection['product']['dmns']['x'] ?? 0),
                        (int) ($connection['product']['dmns']['y'] ?? 0),
                        (int) ($connection['product']['dmns']['z'] ?? 0),
                        count($connection['order_edge_supply']),
                        avia: true
                    );
                } catch (Exception $e) {
                    $connection['delivery']['avia']['error'] = true;

                    // var_dump($e->getMessage());
                    // var_dump($e->getTrace());
                    // var_dump($e->getFile()); die;

                    // var_dump(json_decode($e->getMessage(), true)['errors']); die;
                }

                // Запись цены (цена поставки + цена доставки + наша наценка)
                $connection['cost']['avia'] = ($cost['ЦенаЗаЕдиницу'] ?? $connection['supply']->onec['Цены']['Цена']['ЦенаЗаЕдиницу']) + ($connection['delivery']['avia']['price']['all'] ?? $connection['delivery']['avia']['price']['one'] ?? 0) + ($settings['increase'] ?? 0) ?? 0;
            }

            // Запись валюты
            $connection['currency'] = $cost['Валюта'];
        }

        return $connections;
    }

    /**
     * @return DocumentInterface[]
     */
    public static function findDocuments1c(): ?array
    {
        // yii::$app->on(ApiController::EVENT_AFTER_EXPORT_ORDERS, self::afterExport1c());

        $orders = self::searchByEdge(
            from: 'account',
            to: 'order',
            edge: 'account_edge_order',
            direction: 'INBOUND',
            subquery_where: 'account_edge_order.type == "processed"',
            where: ['sync' => false]
        );

        foreach ($orders as &$order) {
            // Перебор заказов

            // Запись о том, что синхронизация проведена
            $order->sync = true;

            $order->update();
        }

        return $orders;
    }

    /**
     * @return OfferInterface[]
     */
    public function getOffers1c(): mixed
    {
        // Инициализация
        $supplies = [];

        foreach ($this->jrnl as $key => $jrnl) {
            // Перебор журнала

            // if (isset($supplies[$key]) && $supplies[$key]['id'] !== $jrnl['target']) {
            //     // Запись уже существует и идентификаторы не совпадают

            //     $key .= '_du'

            //     // Реинициализация
            //     $supplies[$key]['id'] = $jrnl['target'];
            // } else {
            //     // Инициализация
            // }


            if (($jrnl['action'] ?? null) === 'write') {
                // Найдено событие записи товара к заказу

                $supplies[$key]['id'] = $jrnl['target'];

                $supplies[$key]['amount'] ?? $supplies[$key]['amount'] = 0;

                ++$supplies[$key]['amount'];
            } else if (($jrnl['action'] ?? null) === 'delete') {
                // Найдено событие удаления товара из заказа

                $supplies[$key]['id'] = $jrnl['target'];

                $supplies[$key]['amount'] ?? $supplies[$key]['amount'] = 0;

                --$supplies[$key]['amount'];
            }
        }

        // file_put_contents('supplies.txt', print_r($supplies, true));

        if (count($supplies) > 0) {
            // Поставки были записаны

            // Инициализация
            $supplies_buffer = [];

            foreach ($supplies as $id => $supply) {
                // Перебор поставок

                if ($supply['amount'] < 1) {
                    continue;
                }

                if ($response = Supply::searchById($supply['id'])) {
                    // Поставка найдена в базе данных

                    $supplies_buffer[] = $response;
                }
            }

            return $supplies_buffer;
        }
        // file_put_contents('AAAAAAAAAAAAAAAAAAA.txt', print_r(1, true));

        return [];
    }

    /**
     * Неизвестно для чего
     */
    public function getRequisites1c(): void
    {
        // return true;
    }

    /**
     * Получить контрагента (пользователя от лица которого происходит операция)
     *
     * @return PartnerInterface
     */
    public function getPartner1c(): Account
    {
        return yii::$app->user->identity ?? throw new Exception('Не удалось идентифицировать пользователя');
    }

    public function getExportFields1c($context = null)
    {
        return [];
    }

    /**
     * Возвращаем имя поля в базе данных, в котором хранится ID из 1с
     *
     * @return string
     */
    public static function getIdFieldName1c()
    {
        return 'ocid';
    }

    public function setRaw1cData($cml, $object): void
    {
    }

    protected static function afterExport1c(): void
    {
        file_put_contents('afterExport1c.txt', print_r(1, true));
    }

    /**
     * Отправка уведомления
     */
    public static function notification(string $text, string|null $type = Notification::TYPE_NOTICE): Notification|array|null
    {
        // Отправка
        return Notification::_write($text, type: $type);
    }
}
